// Device-specific include file.
#include <p33fxxxx.h>

// Standard C library headers.
#include <string.h>
#include <stdlib.h>

// LCD functions.
#include "lcd.h"

// Stores a 320x240 monochrome bitmap.
unsigned char __attribute__((far)) lcd_frame_buffer[9600];

// Keeps track of the current row and column when retracing the LCD.
volatile unsigned int lcd_scan_row = 0;
volatile unsigned int lcd_scan_column = 0;

// Points to the next byte of the frame buffer to output when retracing the LCD.
unsigned char* lcd_scan_ptr = lcd_frame_buffer;

// Initialises the LCD.
void lcd_init(void) {

	// Make the LCD data port (A) an output port with all pins low.
	LATA = 0;
	TRISA = 0;

	// Make the LCD data port (B) an output port with all pins low.
	LATB = 0;
	TRISB = 0;

	// Configure timer 1 to trigger interrupts to constantly retrace the LCD.
	T1CONbits.TCKPS = 0b00; // No prescaler.
	PR1 = ((unsigned long)FCY / (320 * (240 / 4)) / LCD_FPS); // Trigger the interrupt when the counter reaches this value.
	
	IFS0bits.T1IF = 0; // Clear the interrupt flag.
	IEC0bits.T1IE = 1; // Enable the interrupt.
	
	TMR1 = 0; // Reset the timer.
	T1CONbits.TON = 1; // Switch on the timer.

	// Switch on the LCD.
	LATB |= (1 << 11);

}

// Clear the LCD.
void lcd_clear(void) {
	memset(lcd_frame_buffer, 0, 9600);
}

// Set a pixel on the LCD.
void lcd_set_pixel(int x, int y) {
	if (x >= 0 && x < 320 && y >= 0 && y < 240) {
		lcd_frame_buffer[y * 40 + x / 8] |= (0x80 >> (x & 7));
	}
}

// Clear a pixel on the LCD.
void lcd_clear_pixel(int x, int y) {
	if (x >= 0 && x < 320 && y >= 0 && y < 240) {
		lcd_frame_buffer[y * 40 + x / 8] &= ~(0x80 >> (x & 7));
	}
}

// Invert a pixel on the LCD.
void lcd_invert_pixel(int x, int y) {
	if (x >= 0 && x < 320 && y >= 0 && y < 240) {
		lcd_frame_buffer[y * 40 + x / 8] ^= (0x80 >> (x & 7));
	}
}

// Draw a pixel on the LCD using the current plotting mode.
void (*lcd_draw_pixel)(int, int) = &lcd_set_pixel;

// Set the current plotting mode.
void lcd_set_plot_mode(enum LCD_PLOT_MODE mode) {
	switch (mode) {
		case LCD_PLOT_MODE_SET:
			lcd_draw_pixel = &lcd_set_pixel;
			break;
		case LCD_PLOT_MODE_CLEAR:
			lcd_draw_pixel = &lcd_clear_pixel;
			break;
		case LCD_PLOT_MODE_INVERT:
			lcd_draw_pixel = &lcd_invert_pixel;
			break;

	}
}

// Draw a line on the LCD.
// Adapted from http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
void lcd_draw_line(int x0, int y0, int x1, int y1) {
	int deltax, deltay;
	int error, ystep;
	int x, y;
	unsigned char steep = abs(y1 - y0) > abs(x1 - x0);
	if (steep) {
		int s;
		s = x0; x0 = y0; y0 = s;
		s = x1; x1 = y1; y1 = s;
	}
	if (x0 > x1) {
		int s;
		s = x0; x0 = x1; x1 = s;
		s = y0; y0 = y1; y1 = s;
	}
	deltax = x1 - x0;
	deltay = abs(y1 - y0);
	error = deltax / 2;
	y = y0;
	if (y0 < y1) {
		ystep = 1;
	} else {
		ystep = -1;
	}
	for (x = x0; x <= x1; ++x) {
		if (steep) {
			lcd_draw_pixel(y, x);
		} else {
			lcd_draw_pixel(x, y);
		}
		error = error - deltay;
		if (error < 0) {
			y = y + ystep;
			error = error + deltax;
		}
	}
}

// Draw a circle on the LCD.
// Adapted from http://en.wikipedia.org/wiki/Midpoint_circle_algorithm
void lcd_draw_circle(int x0, int y0, int radius) {
	int f = 1 - radius;
	int ddF_x = 1;
	int ddF_y = -2 * radius;
	int x = 0;
	int y = radius;
	
	lcd_draw_pixel(x0, y0 + radius);
	lcd_draw_pixel(x0, y0 - radius);
	lcd_draw_pixel(x0 + radius, y0);
	lcd_draw_pixel(x0 - radius, y0);
	
	while(x < y) {
		// ddF_x == 2 * x + 1;
		// ddF_y == -2 * y;
		// f == x*x + y*y - radius*radius + 2*x - y + 1;
		if(f >= 0) {
			y--;
			ddF_y += 2;
			f += ddF_y;
		}
		x++;
		ddF_x += 2;
		f += ddF_x;    
		lcd_draw_pixel(x0 + x, y0 + y);
		lcd_draw_pixel(x0 - x, y0 + y);
		lcd_draw_pixel(x0 + x, y0 - y);
		lcd_draw_pixel(x0 - x, y0 - y);
		lcd_draw_pixel(x0 + y, y0 + x);
		lcd_draw_pixel(x0 - y, y0 + x);
		lcd_draw_pixel(x0 + y, y0 - x);
		lcd_draw_pixel(x0 - y, y0 - x);
	}
}
