// Device-specific include file.
#include <p33fxxxx.h>

// Standard C library headers.
#include <math.h>
#include <stdlib.h>
#include <string.h>

// Utility functions.
#include <libpic30.h>

// LCD functions.
#include "lcd.h"

// Configuration.
_FOSC(OSCIOFNC_ON)
_FOSCSEL(FNOSC_FRCPLL)

// Constants.
#define M_PI (3.1415926535897932384626433832795f)

// Types.
typedef struct {
	int x;
	int y;
} point;

// Delay for a particular number of frames.
void delay_frames(int frames) {
	while (frames--) {
		while (!lcd_scan_row);
		while (lcd_scan_row);
	}
}

// Draws a swirly line pattern.
void demo_swirl(void) {

	float angle, sub_angle;
	unsigned int amplitude;
	unsigned int pass;

	lcd_clear();

	for (pass = 0; pass < 2; ++pass) {
		// Draw the pattern.		
		for (amplitude = 0, angle = 0; amplitude < 320; amplitude += 4, angle += 0.08f) {
			for (sub_angle = 0; sub_angle < 2 * M_PI; sub_angle += M_PI * 0.5f) {
				lcd_draw_line(
					160 + amplitude * sin(angle + sub_angle), 120 + amplitude * cos(angle + sub_angle),
					160 + amplitude * sin(angle + sub_angle + M_PI * 0.5f), 120 + amplitude * cos(angle + sub_angle + M_PI * 0.5f)
				);
			}
		}
		if (pass == 0) {
			// Delay for two seconds.
			delay_frames(LCD_FPS * 2);
			// Next loop, we'll clear the screen.
			lcd_set_plot_mode(LCD_PLOT_MODE_CLEAR);
		}
	}

	// Put the plotting mode back the way it was.
	lcd_set_plot_mode(LCD_PLOT_MODE_SET);
}

// Draws a spinning cube.
void demo_cube(void) {
	

	// Source X, Y and Z of each point of the cube.
	int x, y, z;

	// Rotation in X and Y.
	float rx = 0, ry = 0;
	
	// Computed sin/cos of the rotated X and Y.
	float sin_rx, cos_rx, sin_ry, cos_ry;
	
	// Transformed X, Y and Z.
	float tx, ty, tz;

	// Stores the eight projected vertices.
	point vertices[8];

	// There's always a need for one of these.
	unsigned int i;

	// Size of the cube.
	float rs = 0;
	float s;

	unsigned int demo_timer = 315;

	while (demo_timer--) {

		// Calculate the sin/cos of the rotated X and Y.
		sin_rx = sin(rx);
		cos_rx = cos(rx);
		sin_ry = sin(ry);
		cos_ry = cos(ry);

		// Calculate the size of the cube.
		s = 140.0f * (1.0f - cos(rs));
	
		// Calculate each vertex in turn.
		i = 0;
		for (x = -1; x <= +1; x += 2) {
			for (y = -1; y <= +1; y += 2) {
				for (z = -1; z <= +1; z+= 2) {
	
					// Calculate the transformed vertex.
					tx = y * cos_rx - x * sin_rx;
					ty = -x * cos_rx * sin_ry - y * sin_rx * sin_ry - z * cos_ry;
	  				tz = 3 - x * cos_rx * cos_ry - y * sin_rx * cos_ry + z * sin_ry;
	
					// Project and store.
					vertices[i].x = 160 + (int)(tx * s / tz);
					vertices[i].y = 120 + (int)(ty * s / tz);
					
					// Advance to the next vertex.
					++i;
				}
			}
		}
	
		// Wait to return back to first scanline.
		while (lcd_scan_row);

		// Clear the LCD.
		lcd_clear();
		
		// One face.
		lcd_draw_line(vertices[0].x, vertices[0].y, vertices[1].x, vertices[1].y);
		lcd_draw_line(vertices[1].x, vertices[1].y, vertices[3].x, vertices[3].y);
		lcd_draw_line(vertices[3].x, vertices[3].y, vertices[2].x, vertices[2].y);
		lcd_draw_line(vertices[2].x, vertices[2].y, vertices[0].x, vertices[0].y);
	
		// Opposite face.
		lcd_draw_line(vertices[4].x, vertices[4].y, vertices[5].x, vertices[5].y);
		lcd_draw_line(vertices[5].x, vertices[5].y, vertices[7].x, vertices[7].y);
		lcd_draw_line(vertices[7].x, vertices[7].y, vertices[6].x, vertices[6].y);
		lcd_draw_line(vertices[6].x, vertices[6].y, vertices[4].x, vertices[4].y);
	
		// Connecting the two faces.
		for (i = 0; i < 4; ++i) {
			lcd_draw_line(vertices[i].x, vertices[i].y, vertices[i + 4].x, vertices[i + 4].y);
		}

		rs += 0.02f;
		ry += 0.03f;
		rx += 0.05f;
		delay_frames(3);
	}
	lcd_clear();
}

// Draws a Sierpinksi triangle.
void demo_sierpinski(void) {
	
	float x, y, n;
	unsigned int steps = 15000;

	lcd_clear();
	
	x = rand() / (float)RAND_MAX;
	y = rand() / (float)RAND_MAX;
	
	while (steps--) {
		n = rand() / (float)RAND_MAX;
		if (n <= 1.0f / 3.0f) {
			x *= 0.5f;
			y *= 0.5f;
		} else if (n <= 2.0f / 3.0f) {
			x = (0.5f + x) * 0.5f;
			y = (1.0f + y) * 0.5f;
		} else {
			x = (1.0f + x) * 0.5f;
			y *= 0.5f;
		}
		lcd_set_pixel(x * 320, (1.0f - y) * 240);
	}

	delay_frames(LCD_FPS);
}

// Draws a Mandelbrot fractal. Based on code by Ken Silverman.
void demo_mandelbrot(void) {

	// Device (x,y) coordinates.
	int x, y;
	
	// Used to calculate the colour at each point.
	float mx, ox, my, oy, xx, yy;
	
	// Stores the working colour.
	unsigned int c;

	// Used to dither the results. Taken from http://en.wikipedia.org/wiki/Ordered_dithering
	unsigned int threshold_map[4][4] = {
		{  1,  9,  3, 11 },
		{ 13,  5, 15,  7 },
		{  4, 12,  2, 10 },
		{ 16,  8, 14,  6 }
	};

	for (y = 0; y < 240; ++y) {
		for (x = 0; x < 320; ++x) {
	
			// Calculate coordinates of point to sample in the fractal.
			ox = mx = (x - 160) * 1.0f / 100.0f - 0.7f;
			oy = my = (y - 120) * 1.0f / 100.0f;

			// Calculate the colour at the point.
			for (c = 0; c < 16; ++c) {
				xx = mx * mx;
				yy = my * my;
				if (xx + yy >= 4) break;
				my = mx * my * 2.0f + oy;
				mx = xx - yy + ox;
			}
			
			// Colour the pixel.
			if (c + threshold_map[x & 3][y & 3] > 16) {
				lcd_set_pixel(x, y);
			}
		}
	}
}

// Flips the screen upside down four times.
void demo_flip(void) {

	unsigned int top_row, bottom_row, column;
	unsigned char* top_ptr;
	unsigned char* bottom_ptr;
	unsigned int flip_count;
	
	for (flip_count = 0; flip_count < 4; ++flip_count) {
		while (lcd_scan_row);
		for (top_row = 0, bottom_row = 239; top_row < 120; ++top_row, --bottom_row) {
			top_ptr = lcd_frame_buffer + 40 * top_row;
			bottom_ptr = lcd_frame_buffer + 40 * bottom_row;
			for (column = 0; column < 40; ++column) {
				unsigned char s = *top_ptr;
				*(top_ptr++) = *bottom_ptr;
				*(bottom_ptr++) = s;
			}
		}
		delay_frames(LCD_FPS / 2);
	}

}

// Draws a circular wipe pattern on the screen.
void demo_circle_wipe(void) {
	unsigned int r, jiggle;
	for (r = 0; r <= 220; ++r) {
		for (jiggle = 0; jiggle < 2; ++jiggle) {
			lcd_draw_circle(160 + jiggle, 120, r);
		}
		if (r > 20) {
			lcd_set_plot_mode(LCD_PLOT_MODE_CLEAR);
			for (jiggle = 0; jiggle < 2; ++jiggle) {
				lcd_draw_circle(160 + jiggle, 120, r - 20);
			}
			lcd_set_plot_mode(LCD_PLOT_MODE_SET);
		}
	}
}

// Draw a clock hand at a specified angle and with a specified length.
void demo_clock_draw_hand(float angle, int length) {

	int arrowhead_angle_offset;

	// Main part of the hand.
	lcd_draw_line(160, 120, 160 + length * sin(angle), 120 - length * cos(angle));
	
	// Draw the arrowhead.
	for (arrowhead_angle_offset = -1; arrowhead_angle_offset <= +1; arrowhead_angle_offset += 2) {
		lcd_draw_line(
			160 + length * sin(angle),
			120 - length * cos(angle),
			160 + length * 0.85f * sin(angle + arrowhead_angle_offset * 0.1f),
			120 - length * 0.85f * cos(angle + arrowhead_angle_offset * 0.1f)
		);
	}
}

// Draws the clock demo.
void demo_clock(void) {

	int jitter;
	int hand_position;

	lcd_clear();

	// Draw the outer dial.
	for (jitter = -1; jitter <= 0; ++jitter) {
		// Outer ring.
		lcd_draw_circle(160 + jitter, 120, 118);
		lcd_draw_circle(160, 120 + jitter, 118);
		// Inner ring.
		lcd_draw_circle(160 + jitter, 120, 105);
		lcd_draw_circle(160, 120 + jitter, 105);
	}
	
	// Draw the marks around the dial.
	for (hand_position = 0; hand_position < 60; ++hand_position) {
		float tick_length = (hand_position % 5) ? 97 : 90;
		float angle = (hand_position / 30.0f) * M_PI;
		lcd_draw_line(
			160 + 100 * sin(angle),
			120 - 100 * cos(angle),
			160 + tick_length * sin(angle),
			120 - tick_length * cos(angle)
		);
	}

	for (hand_position = 45; hand_position <= 60; ++hand_position) {
		
		// Draw the current hand position.
		demo_clock_draw_hand((hand_position / 30.0f) * M_PI, 88);
		// Delay one second.
		delay_frames(LCD_FPS);

		// Have we finished?
		if (hand_position == 60) break;

		// Erase the old hand before drawing the next one.
		lcd_set_plot_mode(LCD_PLOT_MODE_CLEAR);
		demo_clock_draw_hand((hand_position / 30.0f) * M_PI, 88);
		lcd_set_plot_mode(LCD_PLOT_MODE_SET);
	}
}

// Slides the image off the bottom of the LCD.
void demo_slide(void) {
	int row;
	for (row = 0; row < 240; row += 2) {
		while (lcd_scan_row);
		memmove(lcd_frame_buffer + (row + 2) * 40, lcd_frame_buffer + row * 40, 9600 - (row + 2) * 40);
		memset(lcd_frame_buffer + row * 40, 0, 40 * 2);
	}
}

int main(void) {

	// Configure PLL prescaler, PLL postscaler, PLL divisor
	PLLFBD = 41; // M = 43
	CLKDIVbits.PLLPOST = 0; // N1=2
	CLKDIVbits.PLLPRE = 0; // N2=2

	// Initialise the LCD.
	lcd_init();
	
	// Run the demos.
	demo_clock();
	demo_slide();
	demo_sierpinski();
	demo_flip();
	demo_circle_wipe();
	demo_swirl();
	demo_cube();
	demo_mandelbrot();

	// Infinite loop.
	for (;;);
	return 0;
}

